"""Piper configuration"""
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Mapping, Sequence


class PhonemeType(str, Enum):
    ESPEAK = "espeak"
    TEXT = "text"


@dataclass
class PiperConfig:
    """Piper configuration"""

    num_symbols: int
    """Number of phonemes"""

    num_speakers: int
    """Number of speakers"""

    sample_rate: int
    """Sample rate of output audio"""

    espeak_voice: str
    """Name of espeak-ng voice or alphabet"""

    length_scale: float
    noise_scale: float
    noise_w: float

    phoneme_id_map: Mapping[str, Sequence[int]]
    """Phoneme -> [id,]"""

    phoneme_type: PhonemeType
    """espeak or text"""

    @staticmethod
    def from_dict(config: Dict[str, Any]) -> "PiperConfig":
        inference = config.get("inference", {})

        return PiperConfig(
            num_symbols=config["num_symbols"],
            num_speakers=config["num_speakers"],
            sample_rate=config["audio"]["sample_rate"],
            noise_scale=inference.get("noise_scale", 0.667),
            length_scale=inference.get("length_scale", 1.0),
            noise_w=inference.get("noise_w", 0.8),
            #
            espeak_voice=config["espeak"]["voice"],
            phoneme_id_map=config["phoneme_id_map"],
            phoneme_type=PhonemeType(config.get("phoneme_type", PhonemeType.ESPEAK)),
        )



# Copyright (C) 2025 easyaspi314
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import socketserver
from threading import Thread, Lock
import datetime
import logging
import json
import errno
from dataclasses import dataclass

from websockets import CloseCode
from websockets.sync.server import serve, ServerConnection


import config
import tts
import audio

# e.g. 2025-01-28T19:16:09.449827-05:00
def get_isoformat(time: datetime.datetime = datetime.datetime.now()):
    return time.astimezone().isoformat()

class SpeekabooHandler:
    """
    Base class for UDP and WebSocket handling
    """

    def cmd_speak(self, json_data: dict):
        """
        Speak
        Speak the given message with the provided voice alias.
        Websockets:
        {
            "id": "<id>",
            "request": "Speak",
            "voice": "EventVoice"
            "message": "This is a test message",
            "badWordFilter": true
        }
        UDP:
        {
            "command": "speak",
            "id": "<id>",
            "voice": "<voice alias>",
            "message": "<message>"
        }
        """

        if "message" not in json_data:
            raise ValueError("No message was provided")
        message = str(json_data["message"]).strip()

        if len(message) == 0:
            raise ValueError("Message is empty")
        if "voice" not in json_data:
            raise ValueError("No voice alias was provided")
        voice = str(json_data["voice"])

        if voice == "" or voice not in config.config["voices"]:
            raise ValueError("Voice alias not found")

        speech_id = tts.add(message, voice=voice)
        return {"speechId": speech_id, "text": message, "voiceName": voice, "pitch": 1.0, "volume": 1.0, "rate": 0.0}

    def cmd_stop(self, _json_data: dict):
        """
        Stop
            If TTS is currently speaking, stop _only_ the current speech.
        
        Websockets:
        {
            "id": "<id>"
            "request": "Stop"
        }
        
        UDP:
        {
            "command": "stop"
        } 
        """
        audio.audio.stop_playback()

        return {}

    def cmd_enable(self, _json_data: dict):
        """
        Enable
            Enable the TTS engine

        Websockets:
        {
            "id": "<id>"
            "request": "Disable" (or "Off")
        }
        
        UDP:
        {
            "command": "disable" (or "off")
        }    
        """
        config.enabled = True
        return {}

    def cmd_disable(self, _json_data: dict):
        """
        Disable
            Disable the TTS engine

        Websockets:
        {
            "id": "<id>"
            "request": "Disable" (or "Off")
        }
        
        UDP:
        {
            "command": "disable" (or "off")
        }    
        """
        config.enabled = False
        return {}


    def cmd_pause(self, _json_data: dict | None = None):
        """
        Pause
        Pause the TTS event queue
        
        Websockets:
        {
            "id": "<id>"
            "request": "Pause"
        }
        
        UDP:
        {
            "command": "pause"
        }
        """
        config.paused = True
        return {}

    def cmd_resume(self, _json_data: dict | None = None):
        """
        Resume
        Resume the TTS event queue
        
        Websockets:
        {
            "id": "<id>"
            "request": "Resume"
        }
        
        UDP:
        {
            "command": "resume"
        }
        """
        config.paused = False
        return {}

    def cmd_clear(self, _json_data: dict):
        """
        Clear
        Clear all pending events in the TTS event queue
        
        Websockets:
        {
            "id": "<id>"
            "request": "Clear"
        }
        
        UDP:
        {
            "command": "clear"
        }
        """
        tts.clear()
        return {}


    def cmd_getinfo(self, _json_data: dict):
        """
        Returns version information, required by Streamer.bot.
        Just echo back what Speaker.bot 0.1.4 returns, with a hint
        that this is actually Speekaboo.
        """
        
        return {
            "instanceId": config.GLOBAL_UUID,
            "name": "Speaker.bot",
            "version": "0.1.4",
            "os": "windows",
            "apiVersion": 2,
            "speekabooVersion": "0.2.0"
        }

    
    # List of subscribable events, as reported by Speaker.bot 0.1.4
    subscribable_events = {
        "application":[
            "startedspeaking", # not implemented
            "stoppedspeaking"  # not implemented
        ],
        "texttospeech":[
            "textqueued",      # implemented
            "engineprocessed", # implemented
            "playing",         # implemented
            "finished",        # implemented
            "deleted",         # not implemented
            "error"            # implemented
        ],
        "voicegate":[
            "profileactivated"
        ]
    }
    def cmd_getevents(self, _json_data: dict):
        """
        Undocumented.
        Returns a list of events that you can subscribe to.
        """
        return {
            "events": self.subscribable_events
        }

    def cmd_commands(self, _json_data: dict):
        """
        Returns a list of commands, required by Streamer.bot.
        """
        return { "commands": [*self.commands_websocket] }

    def cmd_getaliases(self, _json_data: dict):
        """
        Undocumented. 
        """
        voices = []
        for voice in config.config["voices"]:
            voices.append({"id": config.config["voices"][voice]["id"], "name": voice,  "voiceCount": 1})
        return { "aliases": voices }

    def cmd_nop(self, _json_data: dict):
        """
        A no-op
        """
        return {}

    def cmd_stub(self, json_data: dict):
        """
        A no-op that logs.
        """

        logging.warning("Calling stub function %s", json_data.get("request", json_data.get("command", "")))
        return {}


    commands_websocket = {
        "Speak": cmd_speak,
        "Pause": cmd_pause,
        "Resume": cmd_resume,
        "Clear": cmd_clear,
        "Stop": cmd_stop,
        "Off": cmd_disable,
        "Disable": cmd_disable,
        "On": cmd_enable,
        "Enable": cmd_enable,
        "Events": cmd_stub,
        "Mode": cmd_stub,
        "GetEvents": cmd_stub,
        "Subscribe": cmd_nop,   # handled specially
        "Unsubscribe": cmd_nop, # handled specially
        "GetInfo": cmd_getinfo,
        "GetAliases": cmd_stub,
        "GetState": cmd_stub,
        "GetVoiceGateProfiles": cmd_stub,
        "ActivateVoiceGateProfile": cmd_stub,
        "Commands": cmd_commands
    }

    commands_udp = {
        "speak": cmd_speak,
        "stop": cmd_stop,
        "enable": cmd_enable,
        "on": cmd_enable,
        "disable": cmd_disable,
        "off": cmd_disable,
        "profile": cmd_stub,
        "pause": cmd_pause,
        "resume": cmd_resume,
        "clear": cmd_clear,
        "events": cmd_stub,
        "reg": cmd_stub,
        "set": cmd_stub,
        "assign": cmd_stub
    }

    def log_server_error(self, err: OSError, protocol: str, addr: str, port: int):

        message = err.strerror
        match err.errno:
            case errno.EADDRINUSE:
                message = "Address in use. Is there another instance running?"
            case errno.EADDRNOTAVAIL:
                message = f"Address {protocol}://{addr}:{port} is unavailable. Check your firewall."
            case -2 | -3:
                message = f"Address {protocol}://{addr}:{port} is invalid."

        config.Event(
            "WebsocketEvent",
            "internal_event",
            "error",
            { "message": message }
        )


@dataclass
class ConnectionInfo:
    """
    Info for an active connection
    """
    websocket: ServerConnection
    subscribed_events: dict

class WSServer(Thread, SpeekabooHandler, config.Observer):
    """
    Websocket server thread handling Speaker.bot requests.
    """


    def do_subscribe(self, json_data: dict, conn_id: str):
        """
        Subscribe
        Subscribes to a list of events (Websockets only)

        Uses the same format as Streamer.bot requests

        Websockets:
        {
            "id": "<id>",
            "request": "Subscribe",
            "events": {
                "groupName|*": [
                    "event1",
                    "event2",
                    "..."
                ]
            }
        }
        """
        if "events" not in json_data:
            raise ValueError("No events provided")
        
        # sorry i'm lazy
        events = json.loads(json.dumps(json_data["events"]).lower())

        if not isinstance(events, dict):
            raise ValueError("Malformed json")

        subscribed = {}
        if conn_id in self.active_connections:
            subscribed = self.active_connections[conn_id].subscribed_events

        # If "*" is supplied, it subscribes to all events
        if "*" in events:
            subscribed = subscribed | self.subscribable_events
        
        else:
            for group in events:
                # Speaker.bot doesn't actually check this 
                # if not group in subscribable_events:
                #     raise ValueError("Unknown group name {}".format(group))

                if group in self.subscribable_events:
                    for event in events[group]:
                        if event == "*":
                            subscribed[group] |= self.subscribable_events[group]
                        else:
                            subscribed[group] |= events[group]


        # Add to the dictionary of subscriptions

        self.active_connections[conn_id].subscribed_events = subscribed

        logging.info("Subscribed %s, current events: %s", conn_id, subscribed)
        return {"events": events}

    def do_unsubscribe(self, json_data: dict, conn_id: str):
        """
        UnSubscribe / Unsubscribe
        Unsubscribes from a list of events (Websockets only)

        Uses the same format as Streamer.bot requests

        Websockets:
        {
            "id": "<id>",
            "request": "UnSubscribe", // or Unsubscribe
            "events": {
                "type|*": [
                    "event1",
                    "event2",
                    "..."
                ],
                ...
            }
        }
        """
        if "events" not in json_data:
            raise ValueError("No events provided")
        
        events = json_data["events"]

        if not isinstance(events, dict):
            raise ValueError("Malformed json")

        subscribed = {}

        if conn_id in self.active_connections:
            subscribed = self.active_connections[conn_id].subscribed_events
        if "*" in events:
            subscribed = {}

        else:
            for group in events:
                # Speaker.bot doesn't actually check this
                # if not group in subscribable_events:
                #     raise ValueError("Unknown group name {}".format(group))

                if group in self.subscribable_events:
                    for i in range(events[group]):
                        if events[group][i] == "*":
                            subscribed[group] =[]
                        elif events[group][i] in subscribed[group]:
                            for j in range(subscribed[group]):
                                if subscribed[group][j] == events[group][i]:
                                    del subscribed[group][j]
                                    break

        logging.debug("Unsubscribed %s, current events: %s", conn_id, subscribed)

        self.active_connections[conn_id].subscribed_events = subscribed
        return {"events":events}

    def handle_event(self, event_source: str, event_type: str, data: dict):
        """
        Handles a subscribable event, triggered by a config.Event
        
        Format:
        {
            "timeStamp": "iso timestamp in local time",
            "event": {"source": "source", "type": "event}
            "data": { ... }
        }
        """
        if event_source == "internal_event":
            return

        response = {}

        response["timeStamp"] = datetime.datetime.now().astimezone().isoformat()
        response["event"] = {
            "source": event_source,
            "type": event_type
        }
        response["data"] = data

        stringified = json.dumps(response)
        logging.info("Sending event %s", stringified)

        for conn_data in self.active_connections.values():
            if event_source.lower() not in conn_data.subscribed_events:
                continue
            if event_type not in conn_data.subscribed_events[event_source]:
                continue
            conn_data.websocket.send(stringified)


    def event_text_queued(self, _message: tts.MessageInfo):
        """
        todo
        """
        raise NotImplementedError()

    def parse_speaker_bot_websocket(self, message: str, conn_id: str) -> str:
        """
        Parses a websocket request.

        Returns the response as a string.
        """
        if len(message) > 10000:
            logging.error("Ignoring overly long message of %i bytes", len(message))
            return '{"error":"message too long"}'

        try:
            json_data = json.loads(message)
        except json.JSONDecodeError:
            logging.error("Failed to parse Websocket message.")
            return '{"error":"malformed command"}'

        if "request" not in json_data or "id" not in json_data:
            logging.error("Missing id or request in command")
            return '{"error":"malformed command"}'

        if not isinstance(json_data["request"], str) or not isinstance(json_data["id"], str):
            logging.error("Invalid json type for request/id")
            return '{"error":"malformed json"}'

        response = {}
        request = json_data["request"]
        try:
            if request == "Subscribe":
                response = self.do_subscribe(json_data, conn_id)
            elif request in ("UnSubscribe", "Unsubscribe"):
                response = self.do_unsubscribe(json_data, conn_id)
            else:
                response = self.commands_websocket.get(request, self.cmd_stub)(self,json_data)

            respjson = json.dumps({ "id": json_data["id"], "status": "ok", "result": response})

            logging.info("Responding %s", respjson)

            return respjson
        except ValueError as e:
            logging.error("Value Error in command: %s", request, exc_info=e)
            return json.dumps({"id": json_data["id"], "status": "error", "error": str(e) })


    def handle_websocket(self, websocket: ServerConnection):
        """
        Handler for a WebSocket connection.
        """

        if self.shutting_down:
            # Deny any new connections.
            websocket.close(CloseCode.GOING_AWAY)
            return

        # Create an ID for self.active_connections, e.g. 127.0.0.1:45748
        conn_id = f"{websocket.remote_address[0]}:{websocket.remote_address[1]}"
        logging.info("New Websocket connection: %s", conn_id)

        # Add it to the active connection info.
        with self.lock:
            if conn_id in self.active_connections:
                logging.warning("Warning: Duplicate connection!!!")

            self.active_connections[conn_id] = ConnectionInfo(websocket, {})

        # Start parsing the messages until the WebSocket closes.
        for message in websocket:
            if not isinstance(message, str):
                logging.warning("Received bytes instead of str, trying to decode")
                try:
                    message = str(message, encoding="utf-8")
                except UnicodeDecodeError:
                    logging.error("Unicode decode error on bytes object")
                    websocket.send('{"error":"malformed command"}')
                    continue

            logging.info("Received WebSocket: %s", message)
            websocket.send(self.parse_speaker_bot_websocket(str(message), conn_id))

        # Delete the connection
        with self.lock:
            del self.active_connections[conn_id]


    def __init__(self, ws_addr: str = config.config["ws_server_addr"], ws_port: int = config.config["ws_server_port"]):
        """
        Constructor
        """
        Thread.__init__(self, name="Websocket Server Thread")
        config.Observer.__init__(self)
        self.observe('WebsocketEvent', self.handle_event)
        self.addr = ws_addr
        self.port = ws_port
        self.server = None
        self.lock = Lock()
        self.shutting_down = False

        self.active_connections: dict[str, ConnectionInfo] = {}


    def run(self):
        """
        Thread entry point for WSServer
        """
        if config.config["ws_server_enabled"]:
            try:
                with serve(self.handle_websocket, self.addr, self.port) as self.server:
                    logging.info("Running Websocket server at ws://%s:%d", self.addr, self.port)
                    config.Event(
                        "WebsocketEvent",
                        "internal_event",
                        "info",
                        { "message": f"Running WebSocket server at ws://{self.addr}:{self.port}."}
                    )
                    self.server.serve_forever()
            except OSError as err:
                logging.error("Error running Websocket server on ws://%s:%d: %s", self.addr, self.port, err)
                self.log_server_error(err, "ws", self.addr, self.port)

                self.server = None

    def stop(self):
        """
        Stops the WebSocket server.
        """
        if not self.is_alive():
            return
        logging.info("Shutting down WSServer...")
        if self.server is not None:
            self.shutting_down = True
            with self.lock:
                if len(self.active_connections) > 0:
                    logging.info("Waiting for connections to close...")
                    # Close all active connections
                    for conn in self.active_connections.values():
                        conn.websocket.close(CloseCode.GOING_AWAY)

                self.active_connections.clear()

            self.server.shutdown()

            self.shutting_down = False
            self.server = None
        self.join()

    def is_running(self) -> bool:
        return self.server is not None

def parse_speaker_bot_udp(thread, message: str):
    """
    Parses a UDP request. 
    """
    if len(message) > 10000:
        logging.error("Ignoring overly long message of %d bytes", len(message))
        return

    json_data: dict = json.loads(message)
    request = json_data.get("command", "")
    SpeekabooHandler.commands_udp.get(request, SpeekabooHandler.cmd_stub)(thread, json_data)

udp_thread = None

class UDPServer(Thread, SpeekabooHandler):

    """
    UDP Server thread handling Speaker.bot requests.
    """
    class UDPHandler(socketserver.BaseRequestHandler):
        """
        socketserver request handler class
        """
        def setup(self):
            pass

        def handle(self):
            """
            Handles a UDP connection.
            """
            self.data: bytes = self.request[0].strip()
            logging.info("Received UDP data: %s", self.data.decode("utf-8"))
            parse_speaker_bot_udp(udp_thread, self.data.decode("utf-8"))


    def __init__(self, udp_addr: str = config.config["udp_server_addr"], udp_port: int = config.config["udp_server_port"]):
        """
        Constructor
        """
        super().__init__(name="UDP Server Thread")
        self.addr = udp_addr
        self.port = udp_port
        self.server = None

    def run(self):
        """
        UDPServer thread entry point
        """
        if config.config["udp_server_enabled"]:
            try:
                with socketserver.UDPServer((self.addr, self.port), UDPServer.UDPHandler) as self.server: # type: ignore
                    logging.info("Running UDP server at udp://%s:%d", self.addr, self.port)
                    config.Event(
                        "WebsocketEvent",
                        "internal_event",
                        "info",
                        { "message": f"Running UDP server at udp://{self.addr}:{self.port}."}
                    )
                    self.server.serve_forever()

            except OSError as err:
                logging.error("Error running UDP server on udp://%s:%d: %s. Is there another instance running?", self.addr, self.port, err)
                self.log_server_error(err, "udp", self.addr, self.port)

                self.server = None

    def stop(self):
        """
        Shuts down the UDP server.
        """
        if not self.is_alive():
            return
        logging.info("Shutting down UDPServer...")
        if self.server is not None:
            self.server.shutdown()
            self.server = None
        self.join()

    def is_running(self) -> bool:
        return self.server is not None

# Global instances
udp_thread = UDPServer()
ws_thread = WSServer()



# Copyright (C) 2025 easyaspi314
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Manages installed Piper voices.

TODO:
 - Allow adding custom voices
 - Do the downloads myself so we can have progress bars and such
 - Put on its own thread
"""

import time
from pathlib import Path
import logging
import threading
import uuid
import json

from piper import download as PiperDownloader

import config
class VoiceManager:
    def __init__(self):
        if config.data_folder is None:
            raise FileNotFoundError("No data folder found!")
        
        voices_file = config.data_folder / "voices.json"
        # Update daily
        needs_update = not voices_file.exists() or time.time() - voices_file.stat().st_mtime > (60 * 60 * 24)
        try:
            self.voices = PiperDownloader.get_voices(config.data_folder, needs_update)
        except IOError:
            self.voices = PiperDownloader.get_voices(config.data_folder, False)
            logging.error("Couldn't parse downloads!")
        self.language = config.config["voice_language"]
        self.threads: dict[str, threading.Thread] = {}


    # wait for cleanup
    def wait_for_downloads(self):
        notified = False
        for thread in self.threads.values():
            if thread.is_alive():
                if not notified:
                    notified = True
                    print("Waiting for downloads to complete...")
                thread.join()
        self.threads.clear()

    def get_downloadable_voices(self):
        return self.voices
        
    def get_voice_path(self, voice: str):
        
        if voice in config.config["additional_voices"]:
            file = config.config["additional_voices"][voice]
            if Path(file).exists() and Path(file + ".json").exists():
                return Path(file)
        
        if config.data_folder is None:
            return None

        try:
            path = PiperDownloader.find_voice(voice, [config.data_folder])
            return path[0]
        except (IOError, ValueError):
            return None


    def is_voice_installed(self, voice: str) -> bool:
        return self.get_voice_path(voice) is not None
    
    def get_voice_size(self, voice: str) -> int:
        size = 0
        for i in self.voices[voice]["files"]:
            size += self.voices[voice]["files"][i]["size_bytes"]

        return size

    def print_all_voices(self):
        for voice in self.voices:
            print(voice)

    def print_all_voices_lang(self, lang: str):
        for voice in self.voices:
            if self.voices[voice]["language"]["code"] == lang:
                try:
                    PiperDownloader.find_voice(voice, [config.data_folder])
                    print("Installed: ", end="")
                except ValueError:
                    pass
                print(voice)

    def download_thread(self, voice: str):
        result = False
        try:
            PiperDownloader.ensure_voice_exists(voice, [config.data_folder], config.data_folder, self.voices)
            result = True
            config.Event("voices_changed", voice, True)
            logging.info("Done downloading %s with result %s", voice, result)
        except Exception as e: # pylint: disable=broad-except
            logging.error("Failed to download %s!", exc_info=e)
            config.Event("voices_changed", voice, False)
            result = False

    def install_voice(self, voice: str) -> None:
        """
        Installs a voice on a separate thread
        """
        if self.is_voice_installed(voice):
            return

        if voice in self.threads and self.threads[voice].is_alive():
            # be patient!!
            return

        self.threads[voice] = threading.Thread(
            target=self.download_thread,
            args=[voice],
            name=f"Download Thread [{voice}]"
        )
        self.threads[voice].start()
    
    def get_all_installed_voices(self):
        if config.data_folder is None:
            return

        for file in config.data_folder.rglob("*.onnx"):
            name = Path(file).stem
            if self.is_voice_installed(name):
                yield (name, file)

        for voice in config.config["additional_voices"]:
            file = config.config["additional_voices"][voice]
            if Path(file).exists() and Path(file + ".json").exists():
                
                yield (voice, Path(file))

    def get_voice_config(self, voice: str) -> dict | None:
        filename = None
        if voice in config.config["additional_voices"]:
            file = config.config["additional_voices"][voice]
            if Path(file).exists() and Path(file + ".json").exists():
                filename = file + ".json"

        if filename is None:
            try:
                _onnx_path, config_path = PiperDownloader.find_voice(voice, [config.data_folder])
                filename = config_path
            except ValueError:
                return None
        try:
            with open(filename, "r", encoding="utf-8") as f:
                return json.load(f)
        except IOError as e:
            logging.error("IO Error opening %s.", filename, exc_info=e)
            return None
        except json.JSONDecodeError as e:
            logging.error("JSON Error opening %s", filename, exc_info=e)
            return None


    def uninstall_voice(self, voice: str):
        if config.data_folder is None:
            return

        try:
            onnx_path, config_path = PiperDownloader.find_voice(voice, [config.data_folder])
            Path(onnx_path).unlink()
            Path(config_path).unlink()
            config.Event("voices_changed", voice, False)
        except IOError as e:
            logging.info("error", exc_info=e)
            pass

    def register_voice(self, voice_path: str):
        if Path(voice_path).exists() and Path(voice_path + ".json").exists():
            voice = Path(voice_path).stem
            config.config["additional_voices"][voice] = voice_path
            config.Event("voices_changed", voice, True)

        else:
            raise ValueError("Could not find config file {}".format(voice_path))

    def deregister_voice(self, voice: str):
        if voice in config.config["additional_voices"]:
            del config.config["additional_voices"][voice]
            config.Event("voices_changed", voice, False)

    def get_used_aliases(self, voice: str) -> list[str]:
        used_aliases: list[str] = []
        for alias in config.config["voices"].values():
            if alias.get("model_name", "") == voice:
                used_aliases.append(voice)
        return used_aliases
    
    def update_alias(self, name: str, voice: str = "", speaker: int|None = None, noise_scale: float|None = None,
                     length_scale: float|None = None, noise_w: float|None = None, sentence_pause: float|None = None, pitch: float|None = None,
                     volume: float|None = None):

        if not name in config.config["voices"]:
            if voice is None:
                raise ValueError("Cannot update/register a new alias without a voice name!")

            config.config["voices"][name] = {
                "model_name": voice,
                "speaker_id": speaker or 0,
                "noise_scale": noise_scale if noise_scale is not None else 0.667,
                "length_scale": length_scale if length_scale is not None else 1.0,
                "noise_w": noise_w if noise_w is not None else 0.8,
                "sentence_pause": sentence_pause if sentence_pause is not None else 0.2, 
                "pitch": pitch if pitch is not None else 1.0,
                "volume": volume if volume is not None else 1.0,
                "uuid": str(uuid.uuid4())
            }
        else:
        
            if voice is not None:
                config.config["voices"][name]["model_name"] = voice
            
            if speaker is not None:
                config.config["voices"][name]["speaker_id"] = speaker

            if noise_scale is not None:
                config.config["voices"][name]["noise_scale"] = noise_scale

            if length_scale is not None:
                config.config["voices"][name]["length_scale"] = length_scale
            
            if noise_w is not None:
                config.config["voices"][name]["noise_w"] = noise_w
            
            if sentence_pause is not None:
                config.config["voices"][name]["sentence_pause"] = sentence_pause

            if volume is not None:
                config.config["voices"][name]["volume"] = volume

            if pitch is not None:
                config.config["voices"][name]["pitch"] = pitch

vm = VoiceManager()



"""Utility for downloading Piper voices."""
import json
import logging
import shutil
from pathlib import Path
from typing import Any, Dict, Iterable, Set, Tuple, Union
from urllib.request import urlopen

from .file_hash import get_file_hash

URL_FORMAT = "https://huggingface.co/rhasspy/piper-voices/resolve/v1.0.0/{file}"

_DIR = Path(__file__).parent
_LOGGER = logging.getLogger(__name__)

_SKIP_FILES = {"MODEL_CARD"}


class VoiceNotFoundError(Exception):
    pass


def get_voices(
    download_dir: Union[str, Path], update_voices: bool = False
) -> Dict[str, Any]:
    """Loads available voices from downloaded or embedded JSON file."""
    download_dir = Path(download_dir)
    voices_download = download_dir / "voices.json"

    if update_voices:
        # Download latest voices.json
        voices_url = URL_FORMAT.format(file="voices.json")
        _LOGGER.debug("Downloading %s to %s", voices_url, voices_download)
        with urlopen(voices_url) as response, open(
            voices_download, "wb"
        ) as download_file:
            shutil.copyfileobj(response, download_file)

    # Prefer downloaded file to embedded
    voices_embedded = _DIR / "voices.json"
    voices_path = voices_download if voices_download.exists() else voices_embedded

    _LOGGER.debug("Loading %s", voices_path)
    with open(voices_path, "r", encoding="utf-8") as voices_file:
        return json.load(voices_file)


def ensure_voice_exists(
    name: str,
    data_dirs: Iterable[Union[str, Path]],
    download_dir: Union[str, Path],
    voices_info: Dict[str, Any],
):
    assert data_dirs, "No data dirs"
    if name not in voices_info:
        raise VoiceNotFoundError(name)

    voice_info = voices_info[name]
    voice_files = voice_info["files"]
    files_to_download: Set[str] = set()

    for data_dir in data_dirs:
        data_dir = Path(data_dir)

        # Check sizes/hashes
        for file_path, file_info in voice_files.items():
            if file_path in files_to_download:
                # Already planning to download
                continue

            file_name = Path(file_path).name
            if file_name in _SKIP_FILES:
                continue

            data_file_path = data_dir / file_name
            _LOGGER.debug("Checking %s", data_file_path)
            if not data_file_path.exists():
                _LOGGER.debug("Missing %s", data_file_path)
                files_to_download.add(file_path)
                continue

            expected_size = file_info["size_bytes"]
            actual_size = data_file_path.stat().st_size
            if expected_size != actual_size:
                _LOGGER.warning(
                    "Wrong size (expected=%s, actual=%s) for %s",
                    expected_size,
                    actual_size,
                    data_file_path,
                )
                files_to_download.add(file_path)
                continue

            expected_hash = file_info["md5_digest"]
            actual_hash = get_file_hash(data_file_path)
            if expected_hash != actual_hash:
                _LOGGER.warning(
                    "Wrong hash (expected=%s, actual=%s) for %s",
                    expected_hash,
                    actual_hash,
                    data_file_path,
                )
                files_to_download.add(file_path)
                continue

    if (not voice_files) and (not files_to_download):
        raise ValueError(f"Unable to find or download voice: {name}")

    # Download missing files
    download_dir = Path(download_dir)

    for file_path in files_to_download:
        file_name = Path(file_path).name
        if file_name in _SKIP_FILES:
            continue

        file_url = URL_FORMAT.format(file=file_path)
        download_file_path = download_dir / file_name
        download_file_path.parent.mkdir(parents=True, exist_ok=True)

        _LOGGER.debug("Downloading %s to %s", file_url, download_file_path)
        with urlopen(file_url) as response, open(
            download_file_path, "wb"
        ) as download_file:
            shutil.copyfileobj(response, download_file)

        _LOGGER.info("Downloaded %s (%s)", download_file_path, file_url)


def find_voice(name: str, data_dirs: Iterable[Union[str, Path]]) -> Tuple[Path, Path]:
    for data_dir in data_dirs:
        data_dir = Path(data_dir)
        onnx_path = data_dir / f"{name}.onnx"
        config_path = data_dir / f"{name}.onnx.json"

        if onnx_path.exists() and config_path.exists():
            return onnx_path, config_path

    raise ValueError(f"Missing files for voice {name}")



import json
import logging
import wave
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import onnxruntime
from piper_phonemize import phonemize_codepoints, phonemize_espeak, tashkeel_run

from .config import PhonemeType, PiperConfig
from .const import BOS, EOS, PAD
from .util import audio_float_to_int16

_LOGGER = logging.getLogger(__name__)


@dataclass
class PiperVoice:
    session: onnxruntime.InferenceSession
    config: PiperConfig

    @staticmethod
    def load(
        model_path: Union[str, Path],
        config_path: Optional[Union[str, Path]] = None,
        use_cuda: bool = False,
    ) -> "PiperVoice":
        """Load an ONNX model and config."""
        if config_path is None:
            config_path = f"{model_path}.json"

        with open(config_path, "r", encoding="utf-8") as config_file:
            config_dict = json.load(config_file)

        providers: List[Union[str, Tuple[str, Dict[str, Any]]]]
        if use_cuda:
            providers = [
                (
                    "CUDAExecutionProvider",
                    {"cudnn_conv_algo_search": "HEURISTIC"},
                )
            ]
        else:
            providers = ["CPUExecutionProvider"]

        return PiperVoice(
            config=PiperConfig.from_dict(config_dict),
            session=onnxruntime.InferenceSession(
                str(model_path),
                sess_options=onnxruntime.SessionOptions(),
                providers=providers,
            ),
        )

    def phonemize(self, text: str) -> List[List[str]]:
        """Text to phonemes grouped by sentence."""
        if self.config.phoneme_type == PhonemeType.ESPEAK:
            if self.config.espeak_voice == "ar":
                # Arabic diacritization
                # https://github.com/mush42/libtashkeel/
                text = tashkeel_run(text)

            return phonemize_espeak(text, self.config.espeak_voice)

        if self.config.phoneme_type == PhonemeType.TEXT:
            return phonemize_codepoints(text)

        raise ValueError(f"Unexpected phoneme type: {self.config.phoneme_type}")

    def phonemes_to_ids(self, phonemes: List[str]) -> List[int]:
        """Phonemes to ids."""
        id_map = self.config.phoneme_id_map
        ids: List[int] = list(id_map[BOS])

        for phoneme in phonemes:
            if phoneme not in id_map:
                _LOGGER.warning("Missing phoneme from id map: %s", phoneme)
                continue

            ids.extend(id_map[phoneme])
            ids.extend(id_map[PAD])

        ids.extend(id_map[EOS])

        return ids

    def synthesize(
        self,
        text: str,
        wav_file: wave.Wave_write,
        speaker_id: Optional[int] = None,
        length_scale: Optional[float] = None,
        noise_scale: Optional[float] = None,
        noise_w: Optional[float] = None,
        sentence_silence: float = 0.0,
    ):
        """Synthesize WAV audio from text."""
        wav_file.setframerate(self.config.sample_rate)
        wav_file.setsampwidth(2)  # 16-bit
        wav_file.setnchannels(1)  # mono

        for audio_bytes in self.synthesize_stream_raw(
            text,
            speaker_id=speaker_id,
            length_scale=length_scale,
            noise_scale=noise_scale,
            noise_w=noise_w,
            sentence_silence=sentence_silence,
        ):
            wav_file.writeframes(audio_bytes)

    def synthesize_stream_raw(
        self,
        text: str,
        speaker_id: Optional[int] = None,
        length_scale: Optional[float] = None,
        noise_scale: Optional[float] = None,
        noise_w: Optional[float] = None,
        sentence_silence: float = 0.0,
    ) -> Iterable[bytes]:
        """Synthesize raw audio per sentence from text."""
        sentence_phonemes = self.phonemize(text)

        # 16-bit mono
        num_silence_samples = int(sentence_silence * self.config.sample_rate)
        silence_bytes = bytes(num_silence_samples * 2)

        for phonemes in sentence_phonemes:
            phoneme_ids = self.phonemes_to_ids(phonemes)
            yield self.synthesize_ids_to_raw(
                phoneme_ids,
                speaker_id=speaker_id,
                length_scale=length_scale,
                noise_scale=noise_scale,
                noise_w=noise_w,
            ) + silence_bytes

    def synthesize_ids_to_raw(
        self,
        phoneme_ids: List[int],
        speaker_id: Optional[int] = None,
        length_scale: Optional[float] = None,
        noise_scale: Optional[float] = None,
        noise_w: Optional[float] = None,
    ) -> bytes:
        """Synthesize raw audio from phoneme ids."""
        if length_scale is None:
            length_scale = self.config.length_scale

        if noise_scale is None:
            noise_scale = self.config.noise_scale

        if noise_w is None:
            noise_w = self.config.noise_w

        phoneme_ids_array = np.expand_dims(np.array(phoneme_ids, dtype=np.int64), 0)
        phoneme_ids_lengths = np.array([phoneme_ids_array.shape[1]], dtype=np.int64)
        scales = np.array(
            [noise_scale, length_scale, noise_w],
            dtype=np.float32,
        )

        args = {
            "input": phoneme_ids_array,
            "input_lengths": phoneme_ids_lengths,
            "scales": scales
        }

        if self.config.num_speakers <= 1:
            speaker_id = None

        if (self.config.num_speakers > 1) and (speaker_id is None):
            # Default speaker
            speaker_id = 0

        if speaker_id is not None:
            sid = np.array([speaker_id], dtype=np.int64)
            args["sid"] = sid

        # Synthesize through Onnx
        audio = self.session.run(None, args, )[0].squeeze((0, 1))
        audio = audio_float_to_int16(audio.squeeze())
        return audio.tobytes()



#!/usr/bin/env python3
import argparse
import io
import logging
import wave
from pathlib import Path
from typing import Any, Dict

from flask import Flask, request

from . import PiperVoice
from .download import ensure_voice_exists, find_voice, get_voices

_LOGGER = logging.getLogger()


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", default="0.0.0.0", help="HTTP server host")
    parser.add_argument("--port", type=int, default=5000, help="HTTP server port")
    #
    parser.add_argument("-m", "--model", required=True, help="Path to Onnx model file")
    parser.add_argument("-c", "--config", help="Path to model config file")
    #
    parser.add_argument("-s", "--speaker", type=int, help="Id of speaker (default: 0)")
    parser.add_argument(
        "--length-scale", "--length_scale", type=float, help="Phoneme length"
    )
    parser.add_argument(
        "--noise-scale", "--noise_scale", type=float, help="Generator noise"
    )
    parser.add_argument(
        "--noise-w", "--noise_w", type=float, help="Phoneme width noise"
    )
    #
    parser.add_argument("--cuda", action="store_true", help="Use GPU")
    #
    parser.add_argument(
        "--sentence-silence",
        "--sentence_silence",
        type=float,
        default=0.0,
        help="Seconds of silence after each sentence",
    )
    #
    parser.add_argument(
        "--data-dir",
        "--data_dir",
        action="append",
        default=[str(Path.cwd())],
        help="Data directory to check for downloaded models (default: current directory)",
    )
    parser.add_argument(
        "--download-dir",
        "--download_dir",
        help="Directory to download voices into (default: first data dir)",
    )
    #
    parser.add_argument(
        "--update-voices",
        action="store_true",
        help="Download latest voices.json during startup",
    )
    #
    parser.add_argument(
        "--debug", action="store_true", help="Print DEBUG messages to console"
    )
    args = parser.parse_args()
    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)
    _LOGGER.debug(args)

    if not args.download_dir:
        # Download to first data directory by default
        args.download_dir = args.data_dir[0]

    # Download voice if file doesn't exist
    model_path = Path(args.model)
    if not model_path.exists():
        # Load voice info
        voices_info = get_voices(args.download_dir, update_voices=args.update_voices)

        # Resolve aliases for backwards compatibility with old voice names
        aliases_info: Dict[str, Any] = {}
        for voice_info in voices_info.values():
            for voice_alias in voice_info.get("aliases", []):
                aliases_info[voice_alias] = {"_is_alias": True, **voice_info}

        voices_info.update(aliases_info)
        ensure_voice_exists(args.model, args.data_dir, args.download_dir, voices_info)
        args.model, args.config = find_voice(args.model, args.data_dir)

    # Load voice
    voice = PiperVoice.load(args.model, config_path=args.config, use_cuda=args.cuda)
    synthesize_args = {
        "speaker_id": args.speaker,
        "length_scale": args.length_scale,
        "noise_scale": args.noise_scale,
        "noise_w": args.noise_w,
        "sentence_silence": args.sentence_silence,
    }

    # Create web server
    app = Flask(__name__)

    @app.route("/", methods=["GET", "POST"])
    def app_synthesize() -> bytes:
        if request.method == "POST":
            text = request.data.decode("utf-8")
        else:
            text = request.args.get("text", "")

        text = text.strip()
        if not text:
            raise ValueError("No text provided")

        _LOGGER.debug("Synthesizing text: %s", text)
        with io.BytesIO() as wav_io:
            with wave.open(wav_io, "wb") as wav_file:
                voice.synthesize(text, wav_file, **synthesize_args)

            return wav_io.getvalue()

    app.run(host=args.host, port=args.port)


if __name__ == "__main__":
    main()



import argparse
import logging
import sys
import time
import wave
from pathlib import Path
from typing import Any, Dict

from . import PiperVoice
from .download import ensure_voice_exists, find_voice, get_voices

_FILE = Path(__file__)
_DIR = _FILE.parent
_LOGGER = logging.getLogger(_FILE.stem)


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("-m", "--model", required=True, help="Path to Onnx model file")
    parser.add_argument("-c", "--config", help="Path to model config file")
    parser.add_argument(
        "-f",
        "--output-file",
        "--output_file",
        help="Path to output WAV file (default: stdout)",
    )
    parser.add_argument(
        "-d",
        "--output-dir",
        "--output_dir",
        help="Path to output directory (default: cwd)",
    )
    parser.add_argument(
        "--output-raw",
        "--output_raw",
        action="store_true",
        help="Stream raw audio to stdout",
    )
    #
    parser.add_argument("-s", "--speaker", type=int, help="Id of speaker (default: 0)")
    parser.add_argument(
        "--length-scale", "--length_scale", type=float, help="Phoneme length"
    )
    parser.add_argument(
        "--noise-scale", "--noise_scale", type=float, help="Generator noise"
    )
    parser.add_argument(
        "--noise-w", "--noise_w", type=float, help="Phoneme width noise"
    )
    #
    parser.add_argument("--cuda", action="store_true", help="Use GPU")
    #
    parser.add_argument(
        "--sentence-silence",
        "--sentence_silence",
        type=float,
        default=0.0,
        help="Seconds of silence after each sentence",
    )
    #
    parser.add_argument(
        "--data-dir",
        "--data_dir",
        action="append",
        default=[str(Path.cwd())],
        help="Data directory to check for downloaded models (default: current directory)",
    )
    parser.add_argument(
        "--download-dir",
        "--download_dir",
        help="Directory to download voices into (default: first data dir)",
    )
    #
    parser.add_argument(
        "--update-voices",
        action="store_true",
        help="Download latest voices.json during startup",
    )
    #
    parser.add_argument(
        "--debug", action="store_true", help="Print DEBUG messages to console"
    )
    args = parser.parse_args()
    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)
    _LOGGER.debug(args)

    if not args.download_dir:
        # Download to first data directory by default
        args.download_dir = args.data_dir[0]

    # Download voice if file doesn't exist
    model_path = Path(args.model)
    if not model_path.exists():
        # Load voice info
        voices_info = get_voices(args.download_dir, update_voices=args.update_voices)

        # Resolve aliases for backwards compatibility with old voice names
        aliases_info: Dict[str, Any] = {}
        for voice_info in voices_info.values():
            for voice_alias in voice_info.get("aliases", []):
                aliases_info[voice_alias] = {"_is_alias": True, **voice_info}

        voices_info.update(aliases_info)
        ensure_voice_exists(args.model, args.data_dir, args.download_dir, voices_info)
        args.model, args.config = find_voice(args.model, args.data_dir)

    # Load voice
    voice = PiperVoice.load(args.model, config_path=args.config, use_cuda=args.cuda)
    synthesize_args = {
        "speaker_id": args.speaker,
        "length_scale": args.length_scale,
        "noise_scale": args.noise_scale,
        "noise_w": args.noise_w,
        "sentence_silence": args.sentence_silence,
    }

    if args.output_raw:
        # Read line-by-line
        for line in sys.stdin:
            line = line.strip()
            if not line:
                continue

            # Write raw audio to stdout as its produced
            audio_stream = voice.synthesize_stream_raw(line, **synthesize_args)
            for audio_bytes in audio_stream:
                sys.stdout.buffer.write(audio_bytes)
                sys.stdout.buffer.flush()
    elif args.output_dir:
        output_dir = Path(args.output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

        # Read line-by-line
        for line in sys.stdin:
            line = line.strip()
            if not line:
                continue

            wav_path = output_dir / f"{time.monotonic_ns()}.wav"
            with wave.open(str(wav_path), "wb") as wav_file:
                voice.synthesize(line, wav_file, **synthesize_args)

            _LOGGER.info("Wrote %s", wav_path)
    else:
        # Read entire input
        text = sys.stdin.read()

        if (not args.output_file) or (args.output_file == "-"):
            # Write to stdout
            with wave.open(sys.stdout.buffer, "wb") as wav_file:
                voice.synthesize(text, wav_file, **synthesize_args)
        else:
            # Write to file
            with wave.open(args.output_file, "wb") as wav_file:
                voice.synthesize(text, wav_file, **synthesize_args)


if __name__ == "__main__":
    main()



# Copyright (C) 2025 easyaspi314
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import time
import threading
import logging
import queue

import miniaudio

import config
import tts

class AudioThread(threading.Thread):
    """
    Handles audio playback.

    Audio playback uses miniaudio.

    This is super messy.
    """
    busy = False

    def __init__(self):
        """
        Constructor
        """
        super().__init__(name="Audio Thread")
        self.queue = queue.Queue()
        self.condition = threading.Condition()
        self.playing = False
        self.running = False
        self.device = None
        self.playback_devices = None


    def initialize(self):

        if self.device is not None:
            self.device.stop()
            self.device.close()
            self.device = None

        self.get_devices()
        assert self.playback_devices is not None

        wanted_device = config.config["output_device"]

        if wanted_device is not None and wanted_device in self.playback_devices:
            output_device = self.playback_devices[wanted_device][1]["id"]
        else:
            if wanted_device is not None:
                logging.warning("Selected audio device %s not found. Choosing the default.", wanted_device)
            output_device = None
        try:
            self.device = miniaudio.PlaybackDevice(
                output_format=miniaudio.SampleFormat.SIGNED16,
                nchannels=1,
                app_name="Speekaboo",
                device_id=output_device
            )
        except miniaudio.MiniaudioError as e:
            config.Event("raise_error", f"MiniAudio error: {e.args[0]}", e)



    def get_devices(self) -> dict:
        """
        Gets a list of devices in a more friendly format than miniaudio:
        """
        if self.playback_devices is None:
            try:
                devices = miniaudio.Devices()
                self.playback_devices = {}
                playbacks = devices.get_playbacks()
                for d in enumerate(playbacks, 1):
                    if d[1]["name"] in self.playback_devices:
                        logging.warning("warning: %s duplicated", d[1]["name"])
                    self.playback_devices[d[1]["name"]] = (d[0], d[1])
            except miniaudio.MiniaudioError as e:
                logging.error("Error getting audio devices.", exc_info=e)
                self.playback_devices = {}

        return self.playback_devices

    def get_sample_rate(self):
        if self.device is None:
            return -1
        return self.device.sample_rate

    def stream_pcm(self, source):
        """
        https://github.com/irmen/pyminiaudio/blob/master/examples/demo3.py
        """
        required_frames = yield b""  # generator initialization
        idx = 0
        while idx < len(source) and config.running and self.playing:
            required_bytes = required_frames * 1 * 2
            sample_data = source[idx:idx+required_bytes]
            idx += required_bytes
            if not sample_data:
                break
            required_frames = yield sample_data



    def stop_playback(self):
        """
        Stops the active speaking voice
        """
        with self.condition:
            if self.playing and self.device is not None and self.device.running:
                self.device.stop()
            self.playing = False
            self.condition.notify_all()


    def stream_end_callback(self):
        with self.condition:
            self.playing = False
            self.condition.notify_all()

    def run(self):
        """
        Audio thread entrypoint

        TODO: this needs some massive cleanup.
        """
        assert self.device is not None

        self.running = True
        while self.running:

            # todo: make this a condition variable
            while self.running and (len(tts._queue) == 0 or self.playing):
                time.sleep(0.5)

            while len(tts._queue) > 0 and not config.paused and self.running:

                message = tts.peek()
                if message is None or message.parsed_data is None:
                    break

                tts.pop()

                config.Event(
                    "WebsocketEvent",
                    "texttospeech", "playing", {
                    "id": message.id,
                    "timestamp": message.timestamp,
                    "text": message.message,
                    "voiceName": config.config["voices"][message.voice].get("model_name", ""), 
                    "voiceEngine": "Speekaboo Piper",
                    "duration": len(message.parsed_data) / self.get_sample_rate() / 2 * 1000,
                    "volume": 1.0,
                    "rate": 0.0
                })
                # https://github.com/irmen/pyminiaudio/blob/master/examples/playcallbacks.py
                stream = self.stream_pcm(message.parsed_data)
                next(stream)
                callbacks_stream = miniaudio.stream_with_callbacks(stream, end_callback=self.stream_end_callback)
                next(callbacks_stream)
                self.playing = True
                self.device.start(callbacks_stream)

                # Wait for playback to finish
                with self.condition:
                    while self.playing and self.running:
                        print("Playing")
                        self.condition.wait()
                
                config.Event(
                    "WebsocketEvent",
                    "texttospeech",
                    "finished", {
                    "id": message.id,
                    "timestamp": message.timestamp,
                    "text": message.message,
                    "voiceName": config.config["voices"][message.voice].get("model_name", ""), 
                    "voiceEngine": "Speekaboo Piper",
                    "duration": len(message.parsed_data) / self.get_sample_rate() / 2 * 1000,
                    "volume": config.config["voices"][message.voice].get("volume", 1.0),
                    "rate": 0.0
                })

        logging.info("Closing audio thread")

    def stop(self):
        if not self.is_alive():
            return
        logging.info("Shutting down audio")
        self.stop_playback()
        self.running = False
        if self.device is not None:
            self.device.close()
        self.join()

audio = AudioThread()


# Copyright (C) 2025 easyaspi314
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import json
import logging
from typing import Callable
import sys
import getpass
import tkinter as tk
from tkinter import messagebox
import threading
from pathlib import Path
from appdirs import AppDirs
import psutil

# Set up some exception handlers. These are handled in gui.poll().
waiting_exceptions = {}

def exception_hook(exc: type, val: BaseException, tb, name="Main thread"):
    """
    Main exception hook
    """
    waiting_exceptions[name] = (exc, val, tb)
    logging.error("Received exception in %s!", name, exc_info=(exc, val, tb))

    censored = str(val).replace(getpass.getuser(), "$USER")
    try:
        messagebox.showerror(
            title="Exception!",
            message=f"Fatal exception in {name}:\n{exc.__name__}(\"{censored}\")"
        )
    except: # pylint:disable=bare-except
        pass
    # os._exit(0)

def thread_exception_hook(args):
    """
    Exception hook for a sub thread.
    """
    name = "Unknown Thread"
    if isinstance(args.thread, threading.Thread) and args.thread.name :
        name = args.thread.name or "Unknown Thread"
    else:
        name = "Unknown Thread"
    exception_hook(args.exc_type, args.exc_value, args.exc_traceback, name)

threading.excepthook = thread_exception_hook

# Global UUID for the program. Definitely generated randomly, definitely has no
# significance whatsoever :)
GLOBAL_UUID = "3243f6a8-885a-308d-3131-98a2e0370734"
running = True
paused = False
enabled = True
dirinfo = AppDirs("Speekaboo")


def try_create_folder(pathname: str|Path) -> Path:
    # Don't show the username on screen
    censored_path = str(pathname).replace(r'getpass.getuser()', "$USER", 1)
    path = Path(pathname)
    if not path.exists():
        try:
            path.mkdir(mode=0o755, parents=True, exist_ok=True)
        except OSError as e:
            messagebox.showerror(message=f"Cannot create folder {censored_path}: {e.strerror}")
            path = None
            raise e

    elif not path.is_dir():
        
        messagebox.showerror(message=f"{censored_path} is not a directory!")
        logging.error("%s is not a directory!", path)
        path = None
        raise OSError("Config folder is not a directory!")

    return path

default_piper_options = {
    "noise_scale": 0.667,      # How varied the model is
    "length_scale": 1.0,       # How fast the TTS talks (lower = faster)
    "noise_w": 0.8,            # How varied the word length is
    "sentence_silence": 0.2,   # How long to pause between sentences
}

system_mem = psutil.virtual_memory().total // (1024 * 1024) # in MiB

config_folder = try_create_folder(dirinfo.user_config_dir)
data_folder = try_create_folder(Path(dirinfo.user_data_dir) / "voicedata")


def load_config():
    """
    Try to load the config file, in either:
    - C:\\Users\\<username>\\AppData\\Local\\Speekaboo\\Speekaboo.json
    - ~/.config/Speekaboo/Speekaboo.json
    - ~/Library/Application Support/Speekaboo/Speekaboo.json
    """
    default_config = {
        "ws_server_enabled": True,                        # Whether to enable the Websocket server
        "ws_server_addr": "127.0.0.1",                    # Address of the Websocket server
        "ws_server_port": 7580,                           # Port for the Websocket server. The documentation erroneously
                                                          #     lists the default port as 7680, but it's really 7580.
        "udp_server_enabled": True,                       # Whether to enable the UDP server
        "udp_server_addr": "0.0.0.0",                     # Address for the UDP server. Not configurable in Speaker.bot.
        "udp_server_port": 6669,                          # Port for the UDP server.
        "voice_language": "en_US",                        # Voice language
        "voices": {

        },
        "additional_voices": {},                          # Map of additional voices {"voice_name": "path/to/file.onnx"}
        "use_cuda": False,                                # Whether to use Cuda
        "output_device": None,                            # Audio output device (null = default)
        "volume": 1.0,                                    # Output volume
        "queue_delay": 0.0,                               # Delay before playing voices (to allow time for moderation)
        "max_words": 25,                                  # Maximum number of words
        "max_memory_usage": min(512, system_mem // 32),   # How much memory in MIB we use before purging old voices. Default to 512 MiB or 1/32 system memory.
        "text_replacement": "filtered"                    # neuro-sama reference :)
    }
    config_file_path = None

    if config_folder is not None:
        config_file_path = config_folder / "Speekaboo.json"


    if config_file_path is not None and config_file_path.is_file():
        try:
            with open(config_file_path, "r", encoding="utf-8") as cfgfile:
                tmpjson = json.load(cfgfile)
                return (config_file_path, default_config | tmpjson)
        except IOError as e:
            logging.error("Error reading config file: %s", e)
        except json.JSONDecodeError as e:
            logging.error("Error parsing config file: %s", e)

        return (None, default_config)

    # assume first use
    return (config_file_path, default_config)



config_file, config = load_config()

def save_config() -> bool:
    """
    Attempts to save the config file.
    
    Returns True on success.
    """

    try:
        if config_folder is not None:
            if "" in config["voices"]:
                del config["voices"][""]

            stringified = json.dumps(config, indent=4)

            with open(config_folder / "Speekaboo.json", "w", encoding="utf-8") as cfgfile:
                cfgfile.write(stringified)

            return True


    except IOError as e:
        logging.error("Error saving config file: %s", e)
    except (ValueError, TypeError, RecursionError) as e:
        logging.error("Error encoding config file! This is a bug!", exc_info=e)

    return False

stdout_handler = logging.StreamHandler(sys.stdout)
handlers = [stdout_handler]
if config_folder:
    handlers.append(logging.FileHandler(config_folder / "Speekaboo.log"))

logging.basicConfig(level=logging.INFO, handlers=handlers)


# https://stackoverflow.com/a/27315715
class Observer():
    """
    Observer base class
    """
    observers = []
    def __init__(self):
        self.observers.append(self)
        self.observables: dict[str, Callable] = {}
    def observe(self, event_name, callback):
        self.observables[event_name] = callback


class Event():
    def __init__(self, name, *args, autofire = True):
        self.name = name
        self.args = args
        if autofire:
            self.fire()
    def fire(self):
        for observer in Observer.observers:
            if self.name in observer.observables:
                observer.observables[self.name](*self.args)



# Copyright (C) 2025 easyaspi314
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import re
import logging
import sys
import math
import signal
from typing import Any, Literal

import tkinter as tk
from tkinter import ttk, messagebox, simpledialog

import tkfilebrowser
import tktooltip
import darkdetect
import sv_ttk

import cuda_available
import onnxruntime as ort

import config


from server import ws_thread, udp_thread
from voice_manager import vm
import audio
import tts



class ExceptionHandlingTk(tk.Tk, config.Observer):
    """
    Tk root window, but with an active exception handler instead of just freezing.
    """
    def report_callback_exception(self, exc, val, tb):
        config.exception_hook(exc, val, tb, "Tkinter callback")

    def raise_error(self, message, info, *_args):
        messagebox.showerror(title="Error", message=message)
        do_close()

    def __init__(self, *args, **kwargs):
        tk.Tk.__init__(self, *args, **kwargs)

        config.Observer.__init__(self)
        self.observe("raise_error", self.raise_error)


window = ExceptionHandlingTk(className="Speekaboo")
window.title("Speekaboo 0.2.0")

window.geometry("800x500")
window.minsize(width=800, height=500)

# Sun Valley theme
# Use darkdetect to match the system theme
sv_ttk.set_theme(darkdetect.theme())

# Give the lists some padding.
style = ttk.Style()
style.configure("Treeview.Heading", padding=(10, 8))
old_height = style.configure("Treeview", "rowheight")
if isinstance(old_height, int):
    style.configure("Treeview", rowheight=old_height + 6)


notebook = ttk.Notebook(window, padding=5)
notebook.pack(expand=True, fill=tk.BOTH)

class LabeledWidget(ttk.Labelframe):
    """
    Wrapper for making widgets with labels
    """
    def __init__(self, parent, label_text: str, widget_type, *args, **kwargs):
        ttk.Labelframe.__init__(self, parent, padding=5, text=label_text, border=0)

        self.widget= widget_type(self, *args, **kwargs)
        self.widget.pack(expand=True, fill=tk.BOTH, padx=5, pady=5)

        self.configure = self.widget.configure # type: ignore
        self.get = self.widget.get        # type: ignore
        self.set = self.widget.set        # type: ignore 

class ToolTip(tktooltip.ToolTip):
    """
    Just a helper so I don't have to type the delay
    """
    def __init__(self, parent, text, *args, delay=1, **kwargs):
        super().__init__(parent, *args, msg=text, delay=delay, **kwargs)

def do_close():
    """
    Shut down the program
    """
    global window # pylint:disable=global-statement

    if not config.running:
        return
    config.running = False

    if window is not None:
        window.destroy()
        window = None

    vm.wait_for_downloads()
    audio.audio.stop()
    ws_thread.stop()
    udp_thread.stop()
    tts.tts_thread.stop()
    config.save_config()

    sys.exit(0)


class MainTab(ttk.Frame, config.Observer):
    """
    Main tab
    
    Contains:
       - Main controls
       - Log
       - Manual textbox
    """
    def __init__(self, parent, *args, **kwargs):
        ttk.Frame.__init__(self, parent, *args, **kwargs)
        config.Observer.__init__(self)
        self.observe("WebsocketEvent", self.handle_event)
        self.observe("aliases_list_updated", self.update_alias)
        row = 0

        scrollable_wrapper = ttk.Panedwindow(self, orient=tk.VERTICAL)
        scrollable_wrapper.grid(row=row, column=0, columnspan=5, sticky=tk.NSEW)        

        log_frame = ttk.Frame(scrollable_wrapper)

        self.log_box = ttk.Treeview(log_frame, selectmode="browse")
        self.log_box.heading("#0", text="Log", anchor=tk.W)


        log_v_scrollbar = ttk.Scrollbar(log_frame, orient=tk.VERTICAL, command=self.log_box.yview)
        log_v_scrollbar.pack(fill=tk.Y, side=tk.RIGHT)

        self.log_box.pack(expand=True, fill=tk.BOTH)

        self.log_box.configure(yscrollcommand=log_v_scrollbar.set)

        scrollable_wrapper.add(log_frame, weight=1)

        queue_frame= ttk.Frame(scrollable_wrapper)


        self.queue_box = ttk.Treeview(queue_frame, selectmode="browse")
        self.queue_box.heading("#0", text="Queue", anchor=tk.W)

        queue_v_scrollbar = ttk.Scrollbar(queue_frame, orient=tk.VERTICAL, command=self.queue_box.yview)
        queue_v_scrollbar.pack(fill=tk.Y, side=tk.RIGHT)

        self.queue_box.pack(expand=True, fill=tk.BOTH)

        self.queue_box.configure(yscrollcommand=queue_v_scrollbar.set)

        scrollable_wrapper.add(queue_frame, weight=1)

        self.grid_rowconfigure(row, weight=1)

        row += 1

        # Skip the temp voice, which has the key of empty string.
        self.alias_select = ttk.Combobox(self, width=1)
        self.update_alias()
        self.alias_select.grid(row=row, column=0, sticky=tk.NSEW, padx=5, pady=5)

        self.manual_text = ttk.Entry(self)
        self.manual_text.grid(row=row, column=1, columnspan=3, sticky=tk.NSEW, padx=5, pady=5)
        self.manual_text.bind("<Return>", self.manual_send)

        self.play_button = ttk.Button(self, text="Speak", command=self.manual_send)
        self.play_button.grid(row=row, column=4, sticky=tk.NSEW, padx=5, pady=5)

        row += 1

        self.stop_button = ttk.Button(self, text="Stop", command=audio.audio.stop_playback)
        self.stop_button.grid(row=row, column=0, padx=5, pady=5, sticky=tk.NSEW)
        ToolTip(self.stop_button, text="Stop the currently playing message")
        self.pause_button_var = tk.StringVar(value="Pause")
        pause_button = ttk.Button(self, textvariable=self.pause_button_var, command=self.pause)
        pause_button.grid(row=row, column=1, padx=5, pady=5, sticky=tk.NSEW)
        ToolTip(pause_button, text="Pause/resume the queue.")
        
        self.disable_button_var= tk.StringVar(value="Disable")
        disable_button = ttk.Button(self, textvariable=self.disable_button_var, command=self.disable)
        disable_button.grid(row=row, column=2, padx=5, pady=5, sticky=tk.NSEW)
        ToolTip(disable_button, text="Enable/Disable queueing and playback")
        self.quit_button = ttk.Button(self, text="Clear",  command=self.clear)
        self.quit_button.grid(row=row, column=3, padx=5, pady=5, sticky=tk.NSEW)
        ToolTip(self.quit_button, text="Clears the queue")
        self.quit_button = ttk.Button(self, text="Quit", command=do_close)
        self.quit_button.grid(row=row, column=4, columnspan=2, padx=5, pady=5, sticky=tk.NSEW)
        ToolTip(self.quit_button, text="Quits the program")

        for i in range(5):
            self.grid_columnconfigure(i, weight=1, uniform="main_buttons")

        self.pack(expand=True, fill="both")

    def pause(self):
        self.pause_button_var.set("Pause" if config.paused else "Resume")

        config.paused = not config.paused

    def disable(self):
        self.disable_button_var.set("Enable" if config.enabled else "Disable")

        config.enabled = not config.enabled

    def clear(self):
        queued = tts.to_list()
        tts.clear()
        for message in queued:
            self.queue_box.delete(message.id)

    def manual_send(self, _event=None):
        """
        Event trigger for a manual TTS request
        """
        if len(config.config["voices"]) == 0:
            messagebox.showerror(parent=window, message="Before you can use the TTS, add a voice in the Manage Voices tab and then create a Voice Alias!")
            return

        msg = self.manual_text.get()
        voice = self.alias_select.get()
        if len(voice) == 0:
            messagebox.showerror(parent=window, message="Please select a voice")
            return
        self.manual_text.delete(0, tk.END)

        tts.add(msg, voice)

    def update_alias(self, *_args):

        aliases = [voice for voice in config.config["voices"] if voice != '']
        flag = len(self.alias_select["values"]) == 0 and len(aliases) > 0
        selected = self.alias_select.get()
        # removed the selected voice!
        if len(aliases) > 0 and selected not in config.config["voices"]:
            flag = True

        self.alias_select["values"] = aliases
        if flag:
            self.alias_select.set(aliases[0])

    def write_to_log(self, message: str):
        self.log_box.insert("", tk.END, text=message)
        # Scroll to the bottom
        self.log_box.yview_moveto(1.0)

    def write_to_queue(self, message: dict):
        self.queue_box.insert('', 0, iid=message["id"], text=message["text"])

        self.queue_box.yview_moveto(0.0)

    def handle_event(self, event_source: str, event_type: str = '', data: dict|None = None):
        print(event_source, event_type, data)
        if window is None or data is None:
            logging.info("Event on dead instance")
            return

        if event_source == "internal_event":
            match event_type:
                case "loading_voice":
                    self.write_to_log(f"Loading voice: {data['voice']}")
                case "loaded_voice":
                    self.write_to_log(f"Loaded voice {data['voice']}, Estimated memory usage: {data['mem']:.1f} MiB")
                case "error":
                    self.write_to_log(f"Error: {data['message']}")
                case "info":
                    self.write_to_log(f"Info: {data['message']}")
        else:
            match event_type:
                case "textqueued":
                    self.write_to_log(f"Queued: {data['text']}")
                    self.write_to_queue(data)
                case "engineprocessed":
                    self.write_to_log(f"Processed: {data['text']}")
                case "playing":
                    self.write_to_log(f"Playing: {data['text']}")
                    self.queue_box.delete(data["id"])
                case "error":
                    self.write_to_log(f"Error: {data['text']}: {data.get('speekaboo_exception', 'Unknown')}")
                    self.queue_box.delete(data["id"])



main_tab = MainTab(window)
notebook.add(main_tab, text="Main")

class VoiceAliasesTab(ttk.Frame):
    """
    Tab to manage voice aliases and custom voices
     _Aliases______________________________________
    |>voice 1<| Name: ____________________________ |
    | voice 2 | Voice: <name> <quality> <speaker>  |
    | voice 3 | Variation: ____                    |
    |         | Speed: ___         etc             |
    |         | Volume:                            |
    |---------'------------------------------------|

    """
    class VoiceConfigFrame(ttk.Frame, config.Observer):
        """
        Frame for voice options
        """
        def __init__(self, parent, *args, **kwargs):
            ttk.Frame.__init__(self, parent, *args, **kwargs)
            config.Observer.__init__(self)
            self.observe("voices_changed", self.update_voices)

            self.voice_config= {}
            self.name_var = tk.StringVar()
            self.voice_var = tk.StringVar()
            self.voice_var.trace_add('write', self.voice_id_callback)

            self.speaker_id_var = tk.IntVar()
            self.variation_var = tk.DoubleVar()
            self.speed_var = tk.DoubleVar()
            self.volume_var = tk.DoubleVar()
            # self.pitch_var = tk.DoubleVar()
            self.length_variation_var = tk.DoubleVar()
            self.voices = []

            row = 0
            ttk.Label(self, text="Name").grid(row=row, column=0, padx=5, pady=5)
            name_entry = ttk.Label(self, textvariable=self.name_var)
            name_entry.grid(row=row, column=1, sticky=tk.NSEW, padx=5, pady=5)
            row += 1
            self.voice_widget = LabeledWidget(self, "Voice", ttk.Combobox, textvariable=self.voice_var, values=())
            self.update_voices()
            self.voice_widget.grid(row=row, column=0, padx=5, pady=5, sticky=tk.NSEW)
            ToolTip(self.voice_widget, text="Selects which voice file (.onnx) to use")

            self.speaker_id = LabeledWidget(self, "Speaker ID", ttk.Spinbox, from_=0, to=0, increment=1, textvariable=self.speaker_id_var)
            self.speaker_id.grid(row=row, column=1, padx=5, pady=5, sticky=tk.NSEW)
            ToolTip(self.speaker_id, text="Some voices contain multiple speakers.")

            row += 1

            variation_input = LabeledWidget(self, "Pitch Variance", ttk.Spinbox, from_=0.0, to=2.0, increment=0.1, textvariable=self.variation_var)
            variation_input.grid(row=row, column=0, padx=5, pady=5, sticky=tk.NSEW)
            ToolTip(variation_input, text="How much variance to put in the voice pitch. Piper arg: noise_scale")

            volume = LabeledWidget(self, "Volume", ttk.Scale, from_=0.0, to=1.0, variable=self.volume_var)
            volume.grid(row=row, column=1, padx=5, pady=5, sticky=tk.NSEW)

            row += 1

            speech_speed = LabeledWidget(self, "Speech speed", ttk.Spinbox, from_=0.001, to=2.0, increment=0.1, textvariable=self.speed_var)
            speech_speed.grid(row=row, column=0, padx=5, pady=5, sticky=tk.NSEW)
            ToolTip(speech_speed, text="Shorter is faster. Controls the length of each phoneme. Piper arg: length_scale")

            noise_w = LabeledWidget(self, "Speech speed variance", ttk.Spinbox, from_=0.001, to=2.0, increment=0.1, textvariable=self.length_variation_var)
            noise_w.grid(row=row, column=1, padx=5, pady=5, sticky=tk.NSEW)
            ToolTip(noise_w, text="How much variance to put into the length of each phoneme. Piper arg: noise_w")

            row += 1
            save_button = ttk.Button(self, text="Save changes", command=self.save_changes)
            save_button.grid(row=row, column=0, padx=5, pady=5, sticky=tk.NSEW)

            test_button = ttk.Button(self, text="Test", command=self.test_voice)
            test_button.grid(row=row, column=1, padx=5, pady=5, sticky=tk.NSEW)

            self.grid_columnconfigure(0, weight=1)
            self.grid_columnconfigure(1, weight=1)

        def load_voice(self, voice: str):
            """
            Loads a voice and populates the variables
            """
            self.tkraise()
            self.name_var.set(voice)
            self.voice_var.set(config.config["voices"][voice].get("model_name", ''))
            self.voice_widget.widget.set(self.voice_var.get())
            self.variation_var.set(config.config["voices"][voice].get("noise_scale", 0.667))
            self.length_variation_var.set(config.config["voices"][voice].get("noise_w", 0.8))
            self.speed_var.set(config.config["voices"][voice].get("length_scale", 1.0))
            self.volume_var.set(config.config["voices"][voice].get("volume", 1.0))
            self.speaker_id.set(config.config["voices"][voice].get("speaker_id", 0))

        def save_voice(self, voice: str):
            """
            Saves the voice
            """
            if self.voice_widget.widget.current() == -1:
                if len(self.voices) == 0:

                    messagebox.showerror(parent=window, message="Please select a voice. Voices can be downloaded in the Manage Voices tab.")
                else:
                    messagebox.showerror(parent=window, message="Please select a voice first.")

                return

            vm.update_alias(
                name= voice,
                voice = self.voices[self.voice_widget.widget.current()],
                speaker = int(self.speaker_id.get() or 0),
                length_scale=self.speed_var.get(),
                noise_scale=self.variation_var.get(),
                noise_w=self.length_variation_var.get(),
                volume=self.volume_var.get()
            )
            logging.info("Saving voice '%s': %s", voice, config.config["voices"][voice])

        def test_voice(self):
            """
            Saves the voice to empty string temp voice
            """
            self.save_voice("")
            tts.add(message="This is a test message", voice="")

        def save_changes(self):
            voice = self.name_var.get()
            self.save_voice(voice)

        def voice_id_callback(self, _var, _index, _mode):
            """
            Triggered when voice_var is changed
            """
            voice = self.voice_var.get()
            if len(voice) == 0:
                return
            
            logging.info("setting %s to %s", self.name_var.get(), voice)

            self.voice_config = vm.get_voice_config(voice) or {"num_speakers": 1}
            num_speaker = self.voice_config["num_speakers"]
            if self.speaker_id_var.get() > num_speaker - 1:
                self.speaker_id_var.set(0)

            vm.update_alias(self.name_var.get(), voice=voice)
            self.speaker_id.widget.configure(to=num_speaker-1)
            self.voice_widget.widget.set(voice)

        def update_voices(self, _voice = "", _installed = False):
            """
            Update the voices list
            """
            self.voices=[]
            tmp = vm.get_all_installed_voices()
            for i, _j in tmp:
                self.voices.append(i)
            self.voices.sort()
            self.voice_widget.widget["values"] = self.voices
            
    def add_alias_callback(self):
        """
        Callback for when the Add alias button is pressed
        """

        # TODO: theming
        newname = simpledialog.askstring(
            prompt="Enter the name of the alias",
            title="New voice alias",
            parent=window
        )

        if newname is None: # user pressed cancel
            return

        newname = newname.strip()

        if newname in config.config["voices"]:
            messagebox.showerror(parent=window, message="Voice already exists.")
            return

        if len(newname) == 0:
            messagebox.showerror(parent=window, message="The name cannot be blank.")
            return

        vm.update_alias(newname)
        self.aliases.insert("", tk.END, iid=newname, text=newname)
        self.aliases.selection_set([newname])

        config.Event("aliases_list_updated")

    def remove_alias_callback(self):
        """
        Removes an alias.
        """
        if len(self.aliases.selection()) != 1:
            messagebox.showerror(parent=window, message="Select a voice first!")
            return
        alias = self.aliases.selection()[0]
        result = messagebox.askyesno(
            parent=window,
            title="Remove alias?",
            message=f"Remove {alias}? This cannot be undone and may cause queued voices to fail."
        )

        if not result:
            return

        del config.config["voices"][alias]

        self.aliases.delete(alias)
        config.Event("aliases_list_updated")
        self.placeholder.tkraise()

    def rename_alias_callback(self):
        if len(self.aliases.selection()) != 1:
            messagebox.showerror(parent=window, message="Select a voice first!")
            return
        alias = self.aliases.selection()[0]

        if alias not in config.config["voices"]:
            messagebox.showerror(parent=window, message="Whoops, can't find that alias. This is a bug!")
            return
        
        result = simpledialog.askstring(
            parent=window,
            title="Rename alias",
            prompt="Enter the new name for this alias. This may cause existing messages to fail.",
            initialvalue=alias
        )

        if result is None:
            return
        result = result.strip()

        if result in config.config["voices"]:
            messagebox.showerror(message="This voice alias already exists.")
            return

        if len(result) == 0:
            messagebox.showerror(message="The name cannot be blank.")
            return

        config.config["voices"][result] = config.config["voices"][alias]
        del config.config["voices"][alias]

        idx = self.aliases.index(alias)

        self.aliases.delete(alias)
        
        self.aliases.insert("", idx, iid=result, text=result)
        self.frame.load_voice(result)
        config.Event("aliases_list_updated")
        
    def select_voice_callback(self, _event):
        for selection in self.aliases.selection():
            self.frame.load_voice(selection)
            break

        

    def __init__(self, parent, *args, **kwargs):
        ttk.Frame.__init__(self, parent, *args, **kwargs)

        self.pack(expand=True, fill=tk.BOTH)

        aliases_frame = ttk.Frame(self)
        self.aliases = ttk.Treeview(aliases_frame, selectmode="browse")
        self.aliases.heading("#0", text="Aliases", anchor=tk.W)

        for voice in config.config["voices"]:
            # Skip the temp voice, which is empty string
            if voice == "":
                continue
            self.aliases.insert("", tk.END, text=voice, iid=voice)

        scrollbar = ttk.Scrollbar(aliases_frame, orient=tk.VERTICAL, command=self.aliases.yview)
        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)

        self.aliases.pack(expand=True, fill=tk.BOTH)

        self.aliases.bind("<<TreeviewSelect>>", self.select_voice_callback)

        self.aliases.configure(yscrollcommand=scrollbar.set)
        aliases_frame.grid(row=0, column=0, padx=5, pady=5)

        self.placeholder = ttk.Label(self, text="Select or create an alias to edit.", anchor=tk.CENTER)
        self.placeholder.configure(justify="center")
        self.placeholder.grid(row=0, column=1, columnspan=2,  sticky=tk.NSEW)
        self.frame = VoiceAliasesTab.VoiceConfigFrame(self)
        self.frame.grid(row=0, column=1, columnspan=2, sticky=tk.NSEW)

        self.placeholder.tkraise()

        self.add_button = ttk.Button(self, text="Add new alias", command=self.add_alias_callback)
        self.add_button.grid(row=1, column=0, padx=5, pady=5, sticky=tk.NSEW)

        self.remove_button = ttk.Button(self, text="Remove selected alias", command=self.remove_alias_callback)
        self.remove_button.grid(row=1, column=1, padx=5, pady=5, sticky=tk.NSEW)

        self.rename_button = ttk.Button(self, text="Rename selected alias", command=self.rename_alias_callback)
        self.rename_button.grid(row=1, column=2, padx=5, pady=5, sticky=tk.NSEW)

        for i in range(3):
            self.grid_columnconfigure(i, weight=1, uniform="voice_aliases")
        self.grid_rowconfigure(0, weight=1)



notebook.add(VoiceAliasesTab(window), text="Voice Aliases")

class DownloadVoicesTab(ttk.Frame, config.Observer):
    """
    Download Voices tab

    Shows a Treeview with all the available voices to download
    """ 
    def set_installed(self, voice: str, installed: bool):
        """
        Observer callback. This may be called after the window is destroyed.
        """

        if window is None:
            return
        # not in list and installing? add it to additional voices
        if not self.voices_list.exists(voice) and installed:
            self.voices_list.insert("/additional_voices", tk.END, iid=voice, text=voice)
            return

        
        item = self.voices_list.item(voice)
        parent = self.voices_list.parent(voice)
        if parent == '/additional_voices' and not installed:
            self.voices_list.delete(voice)
        elif parent != '' and isinstance(item["values"], list):
            item["values"][1] = "Yes" if installed else "No"

            self.voices_list.item(voice, values=item["values"])

    def check_for_alias(self, name: str):
        """
        Checks if a voice is used in an alias, and displays an error if it does
        """
        used_aliases = vm.get_used_aliases(name)
        if len(used_aliases) != 0:
            if len(used_aliases) == 1:
                message = "alias"
            else:
                message = "aliases"

            messagebox.showerror(
                parent=window,
                message=f"This voice model is used by the following {message}: {', '.join(used_aliases)}. Remove the aliases or change the linked voice model first."
            )
            return True

        return False

    def handle_installbutton(self):
        """
        Handles the install button
        """
        if len(self.voices_list.selection()) == 0:
            return

        selection = self.voices_list.selection()[0]
        item = self.voices_list.item(selection)

        # Manually installed voice
        if self.voices_list.parent(selection) == '/additional_voices':
            if self.check_for_alias(selection):
                return
            vm.deregister_voice(selection)
        elif self.voices_list.parent(selection) != '':
            if vm.is_voice_installed(selection):
                if self.check_for_alias(selection):
                    return

                vm.uninstall_voice(selection)
            elif isinstance(item["values"], list):
                item["values"][1] = "Downloading"
                vm.install_voice(selection)

                self.voices_list.item(selection, values=item["values"])

    def handle_addmanualbutton(self):
        """
        Handler for adding manual voices
        """

        real_treeview_insert = ttk.Treeview.insert

        def treeview_insert_patch(self: ttk.Treeview, parent: str, index: int | Literal['end'], iid: str|int|None =None, **kwargs):
            """
            UGLY HACK:
                TkFileBrowser has a bug where if you bookmark a mounted drive on Linux, it will crash
                because it tries to add the same bookmark twice.

                I might fork the old file browser, bugs aside it is so much nicer than native Tk.
            """
            if iid is not None and self.exists(iid):
                return str(iid)

            return real_treeview_insert(self, parent, index, iid, **kwargs)
        # patch treeview
        ttk.Treeview.insert = treeview_insert_patch
        # ask for file
        path = tkfilebrowser.askopenfilename(parent=self, filetypes=[("ONNX voice model (*.onnx)", "*.onnx")])
        # unpatch treeview
        ttk.Treeview.insert = real_treeview_insert

        if len(path) == 0: # user pressed cancel
            return

        try:
            vm.register_voice(str(path))
            logging.info("Registered voice %s", path)

        except ValueError as e:
            messagebox.showerror(
                parent=window,
                message="Could not open voice model. Make sure when installing VOICE.onnx, VOICE.onnx.json is in the same directory."
            )
            logging.error("Failed to find file", exc_info=e)


    def __init__(self, parent, *args, **kwargs):
        ttk.Frame.__init__(self, parent, *args, **kwargs)
        config.Observer.__init__(self)

        self.observe("voices_changed", self.set_installed)
        treeview_container = ttk.Frame(self)
        treeview_container.grid(row=0, column=0, columnspan=2, sticky=tk.NSEW)
        treeview_container.grid_columnconfigure(0, weight=1)
        treeview_container.grid_rowconfigure(0, weight=1)
        self.voices_list = ttk.Treeview(treeview_container, selectmode="browse", columns=("C1", "C2", "C3", "C4"))
        self.voices_list.column("#0", stretch=tk.YES)
        self.voices_list.column("C1", width=70, stretch=tk.NO)
        self.voices_list.column("C2", width=70, stretch=tk.NO)
        self.voices_list.column("C3", width=70, stretch=tk.NO)
        self.voices_list.column("C4", width=80, stretch=tk.NO)
        self.voices_list.heading("#0", text="Name", anchor=tk.W)
        self.voices_list.heading("C1", text="Quality", anchor=tk.W)
        self.voices_list.heading("C2", text="Installed", anchor=tk.W)
        self.voices_list.heading("C3", text="Size", anchor=tk.W)
        self.voices_list.heading("C4", text="Speakers", anchor=tk.W)

        self.voices_list.grid(row=0, column=0, padx=5, pady=5, sticky=tk.NSEW)
        scrollbar = ttk.Scrollbar(treeview_container, orient=tk.VERTICAL, command=self.voices_list.yview)
        scrollbar.grid(row=0, column=1, sticky=tk.NS)
        self.voices_list.configure(yscrollcommand=scrollbar.set)

        self.parse_voices()
        self.installbutton=ttk.Button(self, text="Install/uninstall selected voice", command=self.handle_installbutton)
        self.installbutton.grid(row=1,column=0, padx=5, pady=5, sticky=tk.NSEW)
        self.addmanualbutton= ttk.Button(self, text="Add voice manually from file", command = self.handle_addmanualbutton)
        self.addmanualbutton.grid(row=1, column=1, padx=5, pady=5, sticky=tk.NSEW)
        self.pack(expand=True, fill="y")
        for i in range(2):
            self.grid_columnconfigure(i, weight=1, uniform='install_button')
        self.grid_rowconfigure(0, weight=1)
    # https://stackoverflow.com/a/14822210
    @staticmethod
    def convert_size(size_bytes):
        if size_bytes == 0:
            return "0B"
        size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
        i = int(math.floor(math.log(size_bytes, 1024)))
        p = math.pow(1024, i)
        s = round(size_bytes / p, 1)
        return "%s %s" % (s, size_name[i])
    
    def parse_voices(self):
        # Clear the voices
        if len(self.voices_list.get_children()) > 0:
            self.voices_list.delete(*self.voices_list.get_children())


        manual_voices = config.config["additional_voices"]

        self.voices_list.insert("", 0, iid="/additional_voices", text="Manually installed voices")
        if len(manual_voices) != 0:
            for key in manual_voices:
                self.voices_list.insert("/additional_voices", tk.END, iid=key, text=key)


        downloadable_voices = vm.get_downloadable_voices()

        # First pass: gather all the available languages
        categorized: dict[str, dict] = {}
        language_ids = set()
        for voice, value in downloadable_voices.items():
            code = value["language"]["code"]
            if not code in language_ids:
                categorized[code] = {}

                # example: English (United States)
                friendly_name = f"{value['language']['name_english']} ({value['language']['country_english']})"
                self.voices_list.insert("", "end", iid="/" + code, text=friendly_name)
                language_ids.add(code)
            is_installed = "Yes" if vm.is_voice_installed(voice) else "No"
            friendly_size = self.convert_size(vm.get_voice_size(voice))
            self.voices_list.insert("/" + code, "end", text=value["name"], iid=value["key"],
                                    values=(
                                        value["quality"],
                                        is_installed,
                                        friendly_size,
                                        value["num_speakers"]))

download_voices_tab = DownloadVoicesTab(window)
notebook.add(download_voices_tab, text="Manage Voices")

class ConfigVar(tk.Variable):
    """
    Wrapper for Tk variables to make them automatically write to config.config
    """
    def write_to_config(self, _var, _index, _mode):
        val = self.get()
        if isinstance(config.config[self.key_name], bool):
            val = bool(val)

        config.config[self.key_name] = val

    def __init__(self, parent, key_name, *args, default_value = None,  **kwargs):
        super().__init__(parent, config.config[key_name] or default_value, *args, **kwargs)
        self.key_name = key_name

        self.trace_add("write", self.write_to_config)


class ConfigStringVar(ConfigVar, tk.StringVar):
    """ String version of the ConfigVar """

class ConfigIntVar(ConfigVar, tk.IntVar):
    """ Int version of the ConfigVar """

class ConfigDoubleVar(ConfigVar, tk.DoubleVar):
    """ Float version of the ConfigVar """


class SettingsTab(ttk.Frame):
    """
    Contains various settings for server ports, audio devices, etc. 
    """
    def __init__(self, parent, *args, **kwargs):
        ttk.Frame.__init__(self, parent, *args, **kwargs, padding=5)

        vcmd = (self.register(self.validate_ip_addr), '%P')
        ivcmd = (self.register(self.on_invalid),)

        row = 0
        ttk.Label(self, text="Most changes require a restart.").grid(row=row, column=0, columnspan=4, padx=5, pady=5, sticky=tk.W)
        row += 1
        ttk.Label(self, text="Server").grid(row=row, column=0, columnspan=4, padx=5, pady=5, sticky=tk.W)
        row += 1

        ws_check = ttk.Checkbutton(self, text="WebSocket Server", variable=ConfigIntVar(self, "ws_server_enabled"))
        ws_check.grid(row=row, column=0, columnspan=2, padx=5, pady=5, sticky=tk.W)
        udp_check = ttk.Checkbutton(self, text="UDP Server", variable=ConfigIntVar(self, "udp_server_enabled"))
        udp_check.grid(row=row, column=2, columnspan=2, padx=5, pady=5, sticky=tk.W)
        row += 1
        self.ws_addr_var = ConfigStringVar(self, key_name="ws_server_addr")
        ws_addr = ttk.Entry(self, name="ws", textvariable=self.ws_addr_var,
                            validate='focusout', validatecommand=vcmd, invalidcommand=ivcmd)
        ws_addr.grid(row=row, column=0, padx=5, pady=5, sticky=tk.EW)
        ToolTip(ws_addr, text="IPv4 address for the WebSocket server (default 127.0.0.1)")

        ws_port = ttk.Spinbox(self, from_=0, to=65535, textvariable=ConfigIntVar(self, key_name="ws_server_port"))
        ws_port.grid(row=row, column=1, padx=5, pady=5, sticky=tk.EW)
        ToolTip(ws_port, text="Port for the WebSocket server (default 7580)")

        self.udp_addr_var = ConfigStringVar(self, key_name="udp_server_addr")
        udp_addr = ttk.Entry(self, name="udp", textvariable=self.udp_addr_var,
                            validate='focusout', validatecommand=vcmd, invalidcommand=ivcmd)

        ToolTip(udp_addr, text="IPv4 address for the UDP server (default 0.0.0.0)")
        udp_addr.grid(row=row, column=2, padx=5, pady=5, sticky=tk.EW)
        udp_port = ttk.Spinbox(self, from_=0, to=65535, textvariable=ConfigIntVar(self, key_name="udp_server_port"))
        udp_port.grid(row=row, column=3, padx=5, pady=5, sticky=tk.EW)
        ToolTip(udp_port, text="Port for the WebSocket server (default 6669)")
        row += 1

        ttk.Label(self, text="Voices").grid(row=row, column=0, columnspan=4, padx=5, pady=5, sticky=tk.W)
        row += 1
        mem_limit = LabeledWidget(self, "Voice cache (MiB)", ttk.Spinbox, from_=128, to=config.system_mem // 4, textvariable=ConfigIntVar(self, key_name="max_memory_usage"))
        mem_limit.grid(row=row, column=0, padx=5, pady=5, sticky=tk.EW)

        ToolTip(mem_limit, text="Each voice takes approximately 100 MB in memory, and they take a while to load,"
                                "so Speekaboo will cache the most recently used voices in memory. Note that"
                                "changing the speaker ID is free.")

        word_limit = LabeledWidget(self, "Word limit (0=off)", ttk.Spinbox, from_=0, to=999, textvariable=ConfigIntVar(self, key_name="max_words"))
        word_limit.grid(row=row, column=1, sticky=tk.EW)


        # TODO: Add support for other EPs

        cuda_enabled_state = tk.DISABLED
        if cuda_available.getCudaDeviceCount() == 0:
            cuda_tooltip = "Currently only supported on NVIDIA GPUs"
            config.config["use_cuda"] = False
        elif "CUDAExecutionProvider" not in ort.get_available_providers():
            cuda_tooltip = "Your GPU seems to be supported, but you need to install the CUDA variant of ONNX:\n    python -m pip install onnxruntime-gpu"
            config.config["use_cuda"] = False
        else:
            cuda_enabled_state = tk.NORMAL
            cuda_tooltip = "Uses your GPU to process TTS. May be beneficial if you don't have many CPU cores."


        cudabox = ttk.Checkbutton(self, text="Use CUDA acceleration", state=cuda_enabled_state, variable=ConfigIntVar(self, key_name="use_cuda"))
        cudabox.grid(row=row, column=2, columnspan=2, padx=5, pady=5, sticky=tk.W)
        ToolTip(cudabox, text=cuda_tooltip)
        row += 1

        # not implemented yet
        self.devices_list = ["[Default]"] + list(audio.audio.get_devices())
        self.device_var = tk.StringVar()
        if config.config["output_device"] is None:
            self.device_var.set("[Default]")
        elif config.config["output_device"] not in self.devices_list:
            self.device_var.set(f"(not found) {config.config['output_device']}")
        else:
            self.device_var.set(config.config["output_device"])

        self.device_var.trace_add("write", self.set_audio_device)
        self.device_select = LabeledWidget(self, "Output device", ttk.Combobox, values=self.devices_list, textvariable=self.device_var, state="readonly")

        self.device_select.grid(row=row, column=0, columnspan=4, padx=5, pady=5, sticky=tk.NSEW)


        ToolTip(word_limit, "Prevents playing messages if there are too many words. Set to 0 to disable this limit.")

        for i in range(4):
            self.grid_columnconfigure(i, weight=1, uniform="config")

        self.pack(fill=tk.BOTH, expand=False)

    def set_audio_device(self, _index, _value, _op):
        device = self.device_var.get()
        if device == "[Default]" and self.device_select.widget.current() == 0:
            config.config["output_device"] = None
        else:
            config.config["output_device"] = device
    

    def validate_ip_addr(self, value: str):
        """
        Validates an IPv4 address.

        Adapted from https://www.pythontutorial.net/tkinter/tkinter-validation/
        """
        pattern = r'[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}'
        if re.fullmatch(pattern, value) is None:
            return False
        numbers = value.split('.')
        for num in numbers:
            if int(num) > 255:
                return False

        return True

    def on_invalid(self):
        """
        Show an error message if the data is not valid

        TODO: try to change it back
        """

        messagebox.showerror(parent=window, message="Please enter a valid IPv4 address.")

settings_tab = SettingsTab(window)
notebook.add(settings_tab, text="Settings")

def poll():
    """
    Polling loop to keep the Tk thread happy and to listen for exceptions
    """
    if len(config.waiting_exceptions) > 0:
        # whoops, we have an exception, show it to the user and exit
        for name, exc in config.waiting_exceptions.items():
            messagebox.showerror(
                title="Exception!",
                message=f"Fatal exception in {name}:\n{exc[0].__name__}(\"{exc[1]}\")"
            )
        do_close()
        return

    if config.running and window is not None:
        window.after(100, poll)

# Catch KeyboardInterrupt gracefully
signal.signal(signal.SIGINT, lambda _x, _y: do_close())

# Run cleanup when the user closes the window
window.protocol("WM_DELETE_WINDOW", do_close)

# Run a slow polling loop to make sure that the main thread can process things
window.after(100, poll)

def start_threads():
    # Start worker threads
    audio.audio.initialize()
    ws_thread.start()
    udp_thread.start()
    audio.audio.start()
    tts.tts_thread.start()

window.after(80, start_threads)

window.mainloop()
# unreachable
do_close()



# Copyright (C) 2025 easyaspi314
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from dataclasses import dataclass
import queue
import logging
from pathlib import Path
from collections import deque
import datetime
from datetime import timezone
from threading import Lock, Thread, Condition
import uuid
import json
import math

import numpy as np
import cachetools
from piper import PiperVoice
from piper.util import audio_float_to_int16
import psutil
from miniaudio import convert_frames, SampleFormat

import config
from voice_manager import vm
from audio import audio

@dataclass
class MessageInfo:
    message: str                # Message
    timestamp: str              # Timestamp
    voice: str                  # Voice to use
    skip: bool                  # Whether to skip this entry
    censor: bool                # Whether to censor bad words for future use
    sender: dict                # for future additions
    id: str                     # Unique UUID
    parsed_data: bytearray|None # Parsed TTS data
    def __str__(self):
        return json.dumps(self)

_lock: Lock = Lock()
_parsing_queue: queue.Queue[MessageInfo] = queue.Queue()
_queue: deque[MessageInfo] = deque()
_condition: Condition = Condition()

def add(message: str, voice: str, timestamp: datetime.datetime = datetime.datetime.now(), censor: bool = False):
    message = message.strip()
    if len(message) == 0 or not config.enabled:
        return

    with _condition:
        msg_id = uuid.uuid4()
        msgtoadd = MessageInfo(
            message = message,
            timestamp = timestamp.astimezone(timezone.utc).isoformat().replace("+00:00", "Z"),
            voice = voice,
            skip = False,
            censor = censor,
            sender = {},
            id = str(msg_id),
            parsed_data=None
        )

        _parsing_queue.put(msgtoadd)

        config.Event(
            "WebsocketEvent",
            "texttospeech",
            "textqueued",
            {
                "id": msgtoadd.id,
                "timestamp": msgtoadd.timestamp,
                "text": message,
                "duration": 0.0,
                "engineName": "Speekaboo Piper",
                "voiceName": config.config["voices"][voice]["model_name"],
                "pitch":0.0,
                "volume": 1.0,
                "rate": 0.0
            }
        )
        _condition.notify_all()
        return str(msg_id)
    

def pop() -> MessageInfo|None:
    with _lock:
        if len(_queue) > 0:
            return _queue.pop()
        else:
            return None

def peek(idx: int = 0) -> MessageInfo | None:
    with _lock:
        if len(_queue) > idx:
            return _queue[idx]
        else:
            return None

def num_items() -> int:
    return len(_queue)

def to_list() -> list[MessageInfo]:
    with _lock:
        lst = list(_queue)
        return lst


def toggle_skip(message: MessageInfo):
    with _lock:
        try:
            idx = _queue.index(message)
            _queue[idx].skip = not _queue[idx].skip
        except ValueError:
            pass

def clear():
    with _lock:
        _queue.clear()


@cachetools.cached(cache=cachetools.LRUCache(maxsize = config.config["max_memory_usage"], getsizeof=lambda x: x[1]))
def get_voice_impl(voicepath: Path) -> tuple[PiperVoice, float]: # model, memory usage
    """
    Since these voices can take up a lot of memory but also take a lot of time to load,
    we use cachetools to make a least 
    """
    logging.info("Loading voice %s", voicepath.stem)
    config.Event(
        "WebsocketEvent",
        "internal_event",
        "loading_voice",
        {
            "voice": voicepath.stem
        }
    )
    proc = psutil.Process()
    # estimate memory usage, since python doesn't manage the memory
    start_memory_usage = proc.memory_info().rss

    try:
        voice = PiperVoice.load(voicepath, use_cuda=config.config["use_cuda"])
    except Exception as e: # pylint:disable=broad-exception-caught
        logging.error("Error initializing acceleration, using CPU instead", exc_info=e)
        config.config["use_cuda"] = False
        voice = PiperVoice.load(voicepath, use_cuda=False)

    end_memory_usage = proc.memory_info().rss
    diff_in_mb = (end_memory_usage - start_memory_usage) / (1024.0 * 1024.0)
    logging.info("Estimated memory usage: %.2f MiB", diff_in_mb)
    config.Event(
        "WebsocketEvent",
        "internal_event",
        "loaded_voice",
        {
            "voice": voicepath.stem,
            "mem": diff_in_mb
        }
    )
    return (voice, diff_in_mb)

def get_voice(voicepath: Path) -> PiperVoice:
    return get_voice_impl(voicepath)[0]

class TTSThread(Thread):
    def __init__(self):
        super().__init__(name="TTS Parsing Thread")
        self.queue = queue.Queue()
        self.running = False

    def parse_tts(self, message: MessageInfo):

        try:
            if message.voice not in config.config["voices"]:
                raise ValueError(f"Invalid voice {message.voice}")

            voice_info = config.config["voices"][message.voice]

            if voice_info.get("model_name", "") == "":
                raise ValueError(f"Voice alias {message.voice} doesn't have a name assigned!")

            voice_path = vm.get_voice_path(voice_info["model_name"])
            if voice_path is None:
                raise ValueError(f"Cannot find voice path for {voice_info['model_name']}")

            voice = get_voice(voice_path)

            logging.info("Speaker ID: %d", voice_info.get("speaker_id", 0))

            raw_pcms = []

            volume = voice_info["volume"]

            num_words = len(message.message.split())

            if config.config["max_words"] != 0 and num_words > config.config["max_words"]:
                return bytearray()

            for sentence in voice.synthesize_stream_raw(message.message,
                    speaker_id=voice_info.get("speaker_id", 0),
                    length_scale=voice_info.get("length_scale", 1.0),
                    noise_scale=voice_info.get("noise_scale", 0.667),
                    noise_w=voice_info.get("noise_w", 0.8)

                    ):
                # Adjust the volume
                if abs(volume - 1.0) > 0.01: # volume != 1.0

                    # Convert back to Signed16 (annoyingly, Piper converts from float to int16 beforehand)
                    le16 = np.dtype(np.int16).newbyteorder('<')
                    buf = np.frombuffer(sentence, le16)

                    # Normalize the volume for a more natural curve
                    # https://stackoverflow.com/a/1165188
                    # 32 seems to feel good.
                    normalized_vol = max(0.001, min((math.pow(32.0, volume) - 1) / (32.0 - 1), 1.0))
                    # Multiply by the normalized volume and convert back to LE16
                    buf = audio_float_to_int16(buf, min(32767.0 * normalized_vol, 32767.0)) # piper/util.py
                    # Convert to bytes
                    sentence = buf.tobytes()

                raw_pcms.append(sentence)

            # Join into a raw buffer
            wav_data = b''.join(raw_pcms)

            # Now convert the sample rate to the native rate.
            converted = convert_frames(SampleFormat.SIGNED16,
                                       from_numchannels=1,
                                       from_samplerate=voice.config.sample_rate,
                                       sourcedata=wav_data,
                                       to_fmt = SampleFormat.SIGNED16,
                                       to_numchannels=1,
                                       to_samplerate=audio.get_sample_rate())

            # Get the duration in milliseconds
            duration = round(len(converted) / 2 / audio.get_sample_rate() * 1000, 2)

            # Emit an event to signal that we processed it
            config.Event(
                "WebsocketEvent",
                "texttospeech",
                "engineprocessed",
                {
                    "id": message.id,
                    "timestamp": message.timestamp,
                    "text": message.message,
                    "duration": duration,
                    "engineName": "Speekaboo Piper",
                    "voiceName": voice_info["model_name"],
                    "pitch":0.0,
                    "volume": volume,
                    "rate": 0.0
                }
            )
            return converted

        except Exception as e: # pylint:disable=broad-exception-caught
            logging.error("Exception in parse_tts:", exc_info=e)
            config.Event(
                "WebsocketEvent",
                "texttospeech",
                "error",
                {
                    "id": message.id,
                    "timestamp": message.timestamp,
                    "text": message.message,
                    "duration": 0.0,
                    "engineName": "Speekaboo Piper",
                    "voiceName": message.voice,
                    "pitch":0.0,
                    "volume": 0.0,
                    "rate": 0.0,
                    "speekaboo_exception": f"{e.args[0]}",
                }
            )
            return None


    def stop(self):
        self.running = False
        if not self.is_alive():
            return
        with _condition:
            _condition.notify_all()

        self.join()
        logging.info("Joined TTS thread")

    def run(self):
        self.running = True
        while self.running:
            with _condition:
                while _parsing_queue.empty() and self.running:
                    _condition.wait()
                if self.running:
                    message = _parsing_queue.get()

                    message.parsed_data = self.parse_tts(message)
                    
                    _parsing_queue.task_done()
                    if message.parsed_data is None:
                        continue
                    with _lock:
                        _queue.append(message)

        logging.info("Done running TTS thread")

tts_thread = TTSThread()



import argparse
import hashlib
import json
import sys
from pathlib import Path
from typing import Union


def get_file_hash(path: Union[str, Path], bytes_per_chunk: int = 8192) -> str:
    """Hash a file in chunks using md5."""
    path_hash = hashlib.md5()
    with open(path, "rb") as path_file:
        chunk = path_file.read(bytes_per_chunk)
        while chunk:
            path_hash.update(chunk)
            chunk = path_file.read(bytes_per_chunk)

    return path_hash.hexdigest()


# -----------------------------------------------------------------------------


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("file", nargs="+")
    parser.add_argument("--dir", help="Parent directory")
    args = parser.parse_args()

    if args.dir:
        args.dir = Path(args.dir)

    hashes = {}
    for path_str in args.file:
        path = Path(path_str)
        path_hash = get_file_hash(path)
        if args.dir:
            path = path.relative_to(args.dir)

        hashes[str(path)] = path_hash

    json.dump(hashes, sys.stdout)


if __name__ == "__main__":
    main()
